/*
 * Copyright 2013, Daehwan Kim <infphilo@gmail.com>
 *
 * This file is part of Bowtie 2.
 *
 * Bowtie 2 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Bowtie 2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Bowtie 2.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "edit.h"
#include "splice_site.h"
#include "aligner_report.h"
#include "aligner_result.h"

float donor_prob[4][donor_len] = {
    {0.340f, 0.604f, 0.092f, 0.001f, 0.001f, 0.526f, 0.713f, 0.071f, 0.160f},
    {0.363f, 0.129f, 0.033f, 0.001f, 0.001f, 0.028f, 0.076f, 0.055f, 0.165f},
    {0.183f, 0.125f, 0.803f, 1.000f, 0.001f, 0.419f, 0.118f, 0.814f, 0.209f},
    {0.114f, 0.142f, 0.073f, 0.001f, 1.000f, 0.025f, 0.093f, 0.059f, 0.462f}
};

float acceptor_prob[4][acceptor_len] = {
    {0.090f, 0.084f, 0.075f, 0.068f, 0.076f, 0.080f, 0.097f, 0.092f, 0.076f, 0.078f, 0.237f, 0.042f, 1.000f, 0.001f, 0.239f},
    {0.310f, 0.310f, 0.307f, 0.293f, 0.326f, 0.330f, 0.373f, 0.385f, 0.410f, 0.352f, 0.309f, 0.708f, 0.001f, 0.001f, 0.138f},
    {0.125f, 0.115f, 0.106f, 0.104f, 0.110f, 0.113f, 0.113f, 0.085f, 0.066f, 0.064f, 0.212f, 0.003f, 0.001f, 1.000f, 0.520f},
    {0.463f, 0.440f, 0.470f, 0.494f, 0.471f, 0.463f, 0.408f, 0.429f, 0.445f, 0.504f, 0.240f, 0.246f, 0.001f, 0.001f, 0.104f}
};

float donor_prob_sum[1 << (donor_len << 1)];
float acceptor_prob_sum1[1 << (acceptor_len1 << 1)];
float acceptor_prob_sum2[1 << (acceptor_len2 << 1)];

void init_junction_prob()
{
    for(size_t i = 0; i < donor_len; i++) {
        ASSERT_ONLY(float sum = 0.0f);
        for(size_t j = 0; j < 4; j++) {
            float prob = donor_prob[j][i];
            assert_gt(prob, 0.0f);
            ASSERT_ONLY(sum += prob);
            donor_prob[j][i] = log(prob / background_prob[j]);
        }
        assert_range(0.9f, 1.05f, sum);
    }
    for(size_t i = 0; i < acceptor_len; i++) {
        ASSERT_ONLY(float sum = 0.0f);
        for(size_t j = 0; j < 4; j++) {
            float prob = acceptor_prob[j][i];
            assert_gt(prob, 0.0f);
            ASSERT_ONLY(sum += prob);
            acceptor_prob[j][i] = log(prob / background_prob[j]);
        }
        assert_range(0.9f, 1.05f, sum);
    }
    
    const size_t donor_elms = 1 << (donor_len << 1);
    for(size_t i = 0; i < donor_elms; i++) {
        float sum = 0.0f;
        for(size_t j = 0; j < donor_len; j++) {
            int base = (i >> (j << 1)) & 0x3;
            sum += donor_prob[base][donor_len - j - 1];
        }
        donor_prob_sum[i] = exp(-sum);
    }
    
    const size_t acceptor_elms1 = 1 << (acceptor_len1 << 1);
    for(size_t i = 0; i < acceptor_elms1; i++) {
        float sum = 0.0f;
        for(size_t j = 0; j < acceptor_len1; j++) {
            int base = (i >> (j << 1)) & 0x3;
            sum += acceptor_prob[base][acceptor_len1 - j - 1];
        }
        acceptor_prob_sum1[i] = exp(-sum);
    }
    
    const size_t acceptor_elms2 = 1 << (acceptor_len2 << 1);
    for(size_t i = 0; i < acceptor_elms2; i++) {
        float sum = 0.0f;
        for(size_t j = 0; j < acceptor_len2; j++) {
            int base = (i >> (j << 1)) & 0x3;
            sum += acceptor_prob[base][acceptor_len - j - 1];
        }
        acceptor_prob_sum2[i] = exp(-sum);
    }
}

ostream& operator<<(ostream& out, const SpliceSite& s)
{
	out << s.ref() << "\t"
        << s.left() << "\t"
        << s.right() << "\t"
        << (s.fw() ? "+" : "-") << endl;
	return out;
}

SpliceSiteDB::SpliceSiteDB(
                           const BitPairReference& refs,
                           const EList<string>& refnames,
                           bool threadSafe,
                           bool write,
                           bool read) :
_numRefs(refs.numRefs()),
_refnames(refnames),
_write(write),
_read(read),
_threadSafe(threadSafe)
{
    assert_gt(_numRefs, 0);
    assert_eq(_numRefs, _refnames.size());
    for(uint64_t i = 0; i < _numRefs; i++) {
        _fwIndex.push_back(new RedBlack<SpliceSitePos, uint32_t>(16 << 10, CA_CAT));
        _bwIndex.push_back(new RedBlack<SpliceSitePos, uint32_t>(16 << 10, CA_CAT));
        _pool.expand();
        _spliceSites.expand();
        _mutex.push_back(MUTEX_T());
    }
    
    donorstr.resize(donor_exonic_len + donor_intronic_len);
    acceptorstr.resize(acceptor_intronic_len + acceptor_exonic_len);
}

SpliceSiteDB::~SpliceSiteDB() {
    assert_eq(_fwIndex.size(), _bwIndex.size());
    assert_eq(_fwIndex.size(), _pool.size());
    for(uint64_t i = 0; i < _numRefs; i++) {
        delete _fwIndex[i];
        delete _bwIndex[i];
        
        EList<Pool*>& pool = _pool[i];
        for(size_t j = 0; j < pool.size(); j++) {
            delete pool[j];
        }
    }
}

size_t SpliceSiteDB::size(uint64_t ref) const {
    if(!_read) return 0;
    
    assert_lt(ref, _numRefs);
    assert_lt(ref, _mutex.size());
    assert_lt(ref, _fwIndex.size());
    assert_eq(_fwIndex.size(), _bwIndex.size());
    ThreadSafe t(const_cast<MUTEX_T*>(&_mutex[ref]), _threadSafe && _write);
    return _fwIndex.size();
}

bool SpliceSiteDB::empty(uint64_t ref) const {
    return size(ref) == 0;
}

bool SpliceSiteDB::addSpliceSite(
                                 const Read& rd,
                                 const AlnRes& rs,
                                 uint32_t minAnchorLen)
{
    if(!_write) return false;
    if(rs.trimmed5p(true) + rs.trimmed3p(true) > 0) return false;

    Coord coord = rs.refcoord();
    uint64_t ref = coord.ref();
    assert_lt(ref, _numRefs);
    const EList<Edit>& edits = rs.ned();
    if(!coord.orient()) {
        Edit::invertPoss(const_cast<EList<Edit>&>(edits), rd.length(), false);
	}
    SpliceSitePos ssp;
    uint32_t refoff = coord.off();
    uint32_t leftAnchorLen = 0, rightAnchorLen = 0;
    size_t eidx = 0;
    size_t last_eidx = 0;
    for(size_t i = 0; i < rd.length(); i++, refoff++) {
        while(eidx < edits.size() && edits[eidx].pos == i) {
			if(edits[eidx].isReadGap()) {
                refoff++;
			} else if(edits[eidx].isRefGap()) {
                assert_gt(refoff, 0);
                refoff--;
            }
            if(edits[eidx].isSpliced()) {
                assert_gt(refoff, 0);
                if(ssp.inited()) {
                    assert(edits[last_eidx].isSpliced());
                    assert_lt(edits[last_eidx].pos, edits[eidx].pos);
                    rightAnchorLen = edits[eidx].pos - edits[last_eidx].pos;
                    if(leftAnchorLen >= minAnchorLen && rightAnchorLen >= minAnchorLen) {
                        bool added = false;
                        assert_lt(ref, _mutex.size());
                        ThreadSafe t(&_mutex[ref], _threadSafe && _write);
                        assert_lt(ref, _fwIndex.size());
                        assert(_fwIndex[ref] != NULL);
                        Node *cur = _fwIndex[ref]->add(pool(ref), ssp, &added);
                        if(added) {
                            assert_lt(ref, _spliceSites.size());
                            _spliceSites[ref].expand();
                            _spliceSites[ref].back().init(ssp.ref(), ssp.left(), ssp.right(), ssp.fw(), ssp.canonical());
                            _spliceSites[ref].back()._readid = rd.rdid;
                            assert(cur != NULL);
                            cur->payload = _spliceSites[ref].size() - 1;
                            
                            SpliceSitePos rssp(ssp.ref(), ssp.right(), ssp.left(), ssp.fw(), ssp.canonical());
                            assert_lt(ref, _bwIndex.size());
                            assert(_bwIndex[ref] != NULL);
                            cur = _bwIndex[ref]->add(pool(ref), rssp, &added);
                            assert(added);
                            assert(cur != NULL);
                            cur->payload = _spliceSites[ref].size() - 1;
                            assert_eq(_fwIndex[ref]->size(), _bwIndex[ref]->size());
                        } else {
                            assert(cur != NULL);
                            assert_lt(ref, _spliceSites.size());
                            assert_lt(cur->payload, _spliceSites[ref].size());
                            if(rd.rdid < _spliceSites[ref][cur->payload]._readid) {
                                _spliceSites[ref][cur->payload]._readid = rd.rdid;
                            }
                        }
                    }
                    leftAnchorLen = rightAnchorLen;
                    rightAnchorLen = 0;
                } else {
                    leftAnchorLen = edits[eidx].pos;
                }
                bool fw = (edits[eidx].splDir == EDIT_SPL_FW || edits[eidx].splDir == EDIT_SPL_UNKNOWN);
                bool canonical = (edits[eidx].splDir != EDIT_SPL_UNKNOWN);
                ssp.init(coord.ref(), refoff - 1, refoff + edits[eidx].splLen, fw, canonical);
                refoff += edits[eidx].splLen;
                last_eidx = eidx;
            }
			eidx++;
		}
	}
    if(ssp.inited()) {
        assert(edits[last_eidx].isSpliced());
        assert_lt(edits[last_eidx].pos, rd.length());
        rightAnchorLen = rd.length() - edits[last_eidx].pos;
        if(leftAnchorLen >= minAnchorLen && rightAnchorLen >= minAnchorLen) {
            bool added = false;
            assert_lt(ref, _mutex.size());
            ThreadSafe t(&_mutex[ref], _threadSafe && _write);
            assert_lt(ref, _fwIndex.size());
            assert(_fwIndex[ref] != NULL);
            Node *cur = _fwIndex[ref]->add(pool(ref), ssp, &added);
            if(added) {
                assert_lt(ref, _spliceSites.size());
                _spliceSites[ref].expand();
                _spliceSites[ref].back().init(ssp.ref(), ssp.left(), ssp.right(), ssp.fw(), ssp.canonical());
                _spliceSites[ref].back()._readid = rd.rdid;
                assert(cur != NULL);
                cur->payload = _spliceSites[ref].size() - 1;
                
                SpliceSitePos rssp(ssp.ref(), ssp.right(), ssp.left(), ssp.fw(), ssp.canonical());
                assert_lt(ref, _bwIndex.size());
                assert(_bwIndex[ref] != NULL);
                cur = _bwIndex[ref]->add(pool(ref), rssp, &added);
                assert(added);
                assert(cur != NULL);
                cur->payload = _spliceSites[ref].size() - 1;
                assert_eq(_fwIndex[ref]->size(), _bwIndex[ref]->size());
            } else {
                assert(cur != NULL);
                assert_lt(ref, _spliceSites.size());
                assert_lt(cur->payload, _spliceSites[ref].size());
                if(rd.rdid < _spliceSites[ref][cur->payload]._readid) {
                    _spliceSites[ref][cur->payload]._readid = rd.rdid;
                }
            }
        }
    }
    if(!coord.orient()) {
		Edit::invertPoss(const_cast<EList<Edit>&>(edits), rd.length(), false);
	}

    return true;
}

bool SpliceSiteDB::getSpliceSite(SpliceSite& ss) const
{
    if(!_read) return false;

    uint64_t ref = ss.ref();
    assert_lt(ref, _numRefs);
    assert_lt(ref, _mutex.size());
    ThreadSafe t(const_cast<MUTEX_T*>(&_mutex[ref]), _threadSafe && _write);
    
    assert_lt(ref, _fwIndex.size());
    assert(_fwIndex[ref] != NULL);
    const Node *cur = _fwIndex[ref]->lookup(ss);
    if(cur == NULL) return false;
    assert(cur != NULL);
    assert_lt(ref, _spliceSites.size());
    ss = _spliceSites[ref][cur->payload];
    return true;
}

void SpliceSiteDB::getLeftSpliceSites(uint32_t ref, uint32_t left, uint32_t range, EList<SpliceSite>& spliceSites) const
{
    if(!_read) return;
    
    assert_lt(ref, _numRefs);
    assert_lt(ref, _mutex.size());
    ThreadSafe t(const_cast<MUTEX_T*>(&_mutex[ref]), _threadSafe && _write);
    assert_gt(range, 0);
    assert_geq(left + 1, range);
    assert_lt(ref, _bwIndex.size());
    assert(_bwIndex[ref] != NULL);
    const Node *cur = _bwIndex[ref]->root();
    if(cur != NULL) getSpliceSites_recur(cur, ref, left + 1 - range, left, spliceSites);
}

void SpliceSiteDB::getRightSpliceSites(uint32_t ref, uint32_t right, uint32_t range, EList<SpliceSite>& spliceSites) const
{
    if(!_read) return;
    
    assert_lt(ref, _numRefs);
    assert_lt(ref, _mutex.size());
    ThreadSafe t(const_cast<MUTEX_T*>(&_mutex[ref]), _threadSafe && _write);
    assert_gt(range, 0);
    assert_gt(right + range, range);
    assert_lt(ref, _fwIndex.size());
    assert(_fwIndex[ref] != NULL);
    const Node *cur = _fwIndex[ref]->root();
    if(cur != NULL) getSpliceSites_recur(cur, ref, right, right + range - 1, spliceSites);
    
}

void SpliceSiteDB::getSpliceSites_recur(
                                        const RedBlackNode<SpliceSitePos, uint32_t> *node,
                                        uint32_t ref,
                                        uint32_t left,
                                        uint32_t right,
                                        EList<SpliceSite>& spliceSites) const
{
    assert(node != NULL);
    bool goleft = true, goright = true;
    if((node->key.ref() > ref) ||
       (node->key.ref() == ref && node->key.left() > right)) {
        goright = false;
    }
    
    if((node->key.ref() < ref) ||
       (node->key.ref() == ref && node->key.left() < left)) {
        goleft = false;
    }
    
    if(goleft && node->left != NULL) {
        getSpliceSites_recur(
                             node->left,
                             ref,
                             left,
                             right,
                             spliceSites);
    }
    
    if(node->key.ref() == ref &&
       node->key.left() >= left && node->key.left() <= right) {
        assert_lt(ref, _spliceSites.size());
        assert_lt(node->payload, _spliceSites[ref].size());
#ifndef NDEBUG
        const SpliceSite& ss = _spliceSites[ref][node->payload];
        assert_eq(ss.ref(), node->key.ref());
        assert(ss.left() == node->key.left() ||
               ss.right() == node->key.left());
#endif
        spliceSites.push_back(_spliceSites[ref][node->payload]);
    }
    
    if(goright && node->right != NULL) {
        getSpliceSites_recur(
                             node->right,
                             ref,
                             left,
                             right,
                             spliceSites);
    }
}

void SpliceSiteDB::print(ofstream& out)
{
    for(uint64_t i = 0; i < _fwIndex.size(); i++) {
        assert(_fwIndex[i] != NULL);
        const Node *root = _fwIndex[i]->root();
        if(root != NULL) print_recur(root, out);
    }
}

void SpliceSiteDB::print_recur(
                               const RedBlackNode<SpliceSitePos, uint32_t> *node,
                               ofstream& out)
{
    if(node == NULL) return;
    print_recur(node->left, out);
    const SpliceSitePos& ssp = node->key;
    assert_lt(ssp.ref(), _refnames.size());
    out << _refnames[ssp.ref()] << "\t"
    << ssp.left() << "\t"
    << ssp.right() << "\t"
    << (ssp.canonical() ? (ssp.fw() ? "+" : "-") : ".") << endl;
    print_recur(node->right, out);
}

void SpliceSiteDB::read(ifstream& in, bool novel)
{
    assert_eq(_numRefs, _refnames.size());
    while(!in.eof()) {
        string refname;
        uint32_t left = 0, right = 0;
        char fw = 0;
        in >> refname >> left >> right >> fw;
        uint32_t ref = 0;
        for(; ref < _refnames.size(); ref++) {
            if(_refnames[ref] == refname) break;
        }
        if(ref >= _numRefs) continue;
        assert_lt(ref, _spliceSites.size());
        _spliceSites[ref].expand();
        _spliceSites[ref].back().init(ref, left, right, fw == '+' || fw == '.', fw != '.');
        _spliceSites[ref].back()._fromfile = true;
        assert_gt(_spliceSites[ref].size(), 0);
        
        bool added = false;
        assert_lt(ref, _fwIndex.size());
        assert(_fwIndex[ref] != NULL);
        Node *cur = _fwIndex[ref]->add(pool(ref), _spliceSites[ref].back(), &added);
        assert(added);
        assert(cur != NULL);
        cur->payload = _spliceSites[ref].size() - 1;
        
        added = false;
        SpliceSitePos rssp(ref, right, left, fw == '+' || fw == '.', fw != '.');
        assert_lt(ref, _bwIndex.size());
        assert(_bwIndex[ref] != NULL);
        cur = _bwIndex[ref]->add(pool(ref), rssp, &added);
        assert(added);
        assert(cur != NULL);
        cur->payload = _spliceSites[ref].size() - 1;
    }
}

Pool& SpliceSiteDB::pool(uint64_t ref) {
    assert_lt(ref, _numRefs);
    assert_lt(ref, _pool.size());
    EList<Pool*>& pool = _pool[ref];
    if(pool.size() <= 0 || pool.back()->full()) {
        pool.push_back(new Pool(1 << 20 /* 1MB */, 16 << 10 /* 16KB */, CA_CAT));
    }
    assert(pool.back() != NULL);
    return *pool.back();
}

